@@ -1,7 +1,7 @@
 import numpy
 from numba import prange
 
-from .._common import jitted
+from .._common import jitted, norm2d
 
 Big = 1.0e5
 eps = 1.0e-15
@@ -10,13 +10,13 @@ epsin = 5
 
 @jitted("f8(i4, i4, f8, f8, f8, f8, f8)")
 def t_ana(i, j, dz, dx, zsa, xsa, vzero):
-    """Calculate analytical times in homogenous model."""
+    """Calculate analytical times in homogeneous model."""
     return vzero * ((dz * (i - zsa)) ** 2.0 + (dx * (j - xsa)) ** 2.0) ** 0.5
 
 
 @jitted("UniTuple(f8, 3)(i4, i4, f8, f8, f8, f8, f8)")
 def t_anad(i, j, dz, dx, zsa, xsa, vzero):
-    """Calculate analytical times in homogenous model and derivatives of times."""
+    """Calculate analytical times in homogeneous model and derivatives of times."""
     t = t_ana(i, j, dz, dx, zsa, xsa, vzero)
 
     if t > 0.0:
@@ -68,11 +68,11 @@ def delta(
 
 
 @jitted(
-    "void(f8[:, :], f8[:, :, :], f8[:, :], UniTuple(f8, 6), f8, f8, f8, f8, f8, i4, i4, i4, i4, i4, i4, i4, i4, b1)"
+    "void(f8[:, :], i4[:, :, :], f8[:, :], UniTuple(f8, 6), f8, f8, f8, f8, f8, i4, i4, i4, i4, i4, i4, i4, i4, b1)"
 )
 def sweep(
     tt,
-    ttgrad,
+    ttsgn,
     slow,
     dargs,
     zsi,
@@ -177,18 +177,18 @@ def sweep(
     # Compute gradient according to minimum time direction
     if grad and tt[i, j] != t0:
         if tt[i, j] == t1d1:
-            ttgrad[i, j, 0] = sgntz
-            ttgrad[i, j, 1] = 0.0
+            ttsgn[i, j, 0] = sgntz
+            ttsgn[i, j, 1] = 0
         elif tt[i, j] == t1d2:
-            ttgrad[i, j, 0] = 0.0
-            ttgrad[i, j, 1] = sgntx
+            ttsgn[i, j, 0] = 0
+            ttsgn[i, j, 1] = sgntx
         else:
-            ttgrad[i, j, 0] = sgntz
-            ttgrad[i, j, 1] = sgntx
+            ttsgn[i, j, 0] = sgntz
+            ttsgn[i, j, 1] = sgntx
 
 
-@jitted("void(f8[:, :], f8[:, :, :], f8[:, :], f8, f8, f8, f8, f8, f8, f8, i4, i4, b1)")
-def sweep2d(tt, ttgrad, slow, dz, dx, zsi, xsi, zsa, xsa, vzero, nz, nx, grad):
+@jitted("void(f8[:, :], i4[:, :, :], f8[:, :], f8, f8, f8, f8, f8, f8, f8, i4, i4, b1)")
+def sweep2d(tt, ttsgn, slow, dz, dx, zsi, xsi, zsa, xsa, vzero, nz, nx, grad):
     """Perform one sweeping."""
     dzi = 1.0 / dz
     dxi = 1.0 / dx
@@ -200,7 +200,7 @@ def sweep2d(tt, ttgrad, slow, dz, dx, zsi, xsi, zsa, xsa, vzero, nz, nx, grad):
         for i in range(1, nz):
             sweep(
                 tt,
-                ttgrad,
+                ttsgn,
                 slow,
                 dargs,
                 zsi,
@@ -222,7 +222,7 @@ def sweep2d(tt, ttgrad, slow, dz, dx, zsi, xsi, zsa, xsa, vzero, nz, nx, grad):
         for i in range(nz - 2, -1, -1):
             sweep(
                 tt,
-                ttgrad,
+                ttsgn,
                 slow,
                 dargs,
                 zsi,
@@ -245,7 +245,7 @@ def sweep2d(tt, ttgrad, slow, dz, dx, zsi, xsi, zsa, xsa, vzero, nz, nx, grad):
         for i in range(1, nz):
             sweep(
                 tt,
-                ttgrad,
+                ttsgn,
                 slow,
                 dargs,
                 zsi,
@@ -267,7 +267,7 @@ def sweep2d(tt, ttgrad, slow, dz, dx, zsi, xsi, zsa, xsa, vzero, nz, nx, grad):
         for i in range(nz - 2, -1, -1):
             sweep(
                 tt,
-                ttgrad,
+                ttsgn,
                 slow,
                 dargs,
                 zsi,
@@ -294,8 +294,8 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
     nz, nx = numpy.shape(slow)
 
     # Check inputs
-    condz = 0.0 <= zsrc <= dz * (nz - 1)
-    condx = 0.0 <= xsrc <= dx * (nx - 1)
+    condz = 0.0 <= zsrc <= dz * nz
+    condx = 0.0 <= xsrc <= dx * nx
     if not (condz and condx):
         raise ValueError("source out of bound")
 
@@ -316,11 +316,14 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
     nz += 1
     nx += 1
     tt = numpy.full((nz, nx), Big, dtype=numpy.float64)
-    ttgrad = (
-        numpy.zeros((nz, nx, 2), dtype=numpy.float64)
-        if grad
-        else numpy.empty((0, 0, 0), dtype=numpy.float64)
-    )
+
+    if grad:
+        ttgrad = numpy.zeros((nz, nx, 2), dtype=numpy.float64)
+        ttsgn = numpy.zeros((nz, nx, 2), dtype=numpy.int32)
+
+    else:
+        ttgrad = numpy.empty((0, 0, 0), dtype=numpy.float64)
+        ttsgn = numpy.empty((0, 0, 0), dtype=numpy.int32)
 
     # Do our best to initialize source
     dzu = numpy.abs(zsa - float(zsi))
@@ -402,8 +405,8 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
                 1,
             )
             if grad:
-                ttgrad[zsi + 1, j, 0] = tzc
-                ttgrad[zsi + 1, j, 1] = txc
+                ttsgn[zsi + 1, j, 0] = 1
+                ttsgn[zsi + 1, j, 1] = 1
 
             if dzu > 0.0:
                 dzi = 1.0 / dzu
@@ -428,8 +431,8 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
                     1,
                 )
                 if grad:
-                    ttgrad[zsi, j, 0] = tzc
-                    ttgrad[zsi, j, 1] = txc
+                    ttsgn[zsi, j, 0] = -1
+                    ttsgn[zsi, j, 1] = 1
 
         td[xsi] = vzero * dxw * dx
         for j in range(xsi - 1, -1, -1):
@@ -460,8 +463,8 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
                 -1,
             )
             if grad:
-                ttgrad[zsi + 1, j, 0] = tzc
-                ttgrad[zsi + 1, j, 1] = txc
+                ttsgn[zsi + 1, j, 0] = 1
+                ttsgn[zsi + 1, j, 1] = -1
 
             if dzu > 0.0:
                 dzi = 1.0 / dzu
@@ -488,8 +491,8 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
                     -1,
                 )
                 if grad:
-                    ttgrad[zsi, j, 0] = tzc
-                    ttgrad[zsi, j, 1] = txc
+                    ttsgn[zsi, j, 0] = -1
+                    ttsgn[zsi, j, 1] = -1
 
         dzi = 1.0 / dz
         dz2i = dzi / dz
@@ -523,8 +526,8 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
                 1,
             )
             if grad:
-                ttgrad[i, xsi + 1, 0] = tzc
-                ttgrad[i, xsi + 1, 1] = txc
+                ttsgn[i, xsi + 1, 0] = 1
+                ttsgn[i, xsi + 1, 1] = 1
 
             if dxw > 0.0:
                 dxi = 1.0 / dxw
@@ -549,8 +552,8 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
                     -1,
                 )
                 if grad:
-                    ttgrad[i, xsi, 0] = tzc
-                    ttgrad[i, xsi, 1] = txc
+                    ttsgn[i, xsi, 0] = 1
+                    ttsgn[i, xsi, 1] = -1
 
         td[zsi] = vzero * dzu * dz
         for i in range(zsi - 1, -1, -1):
@@ -581,8 +584,8 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
                 1,
             )
             if grad:
-                ttgrad[i, xsi + 1, 0] = tzc
-                ttgrad[i, xsi + 1, 1] = txc
+                ttsgn[i, xsi + 1, 0] = -1
+                ttsgn[i, xsi + 1, 1] = 1
 
             if dxw > 0.0:
                 dxi = 1.0 / dxw
@@ -607,28 +610,34 @@ def fteik2d(slow, dz, dx, zsrc, xsrc, nsweep=2, grad=False):
                     -1,
                 )
                 if grad:
-                    ttgrad[i, xsi, 0] = tzc
-                    ttgrad[i, xsi, 1] = txc
+                    ttsgn[i, xsi, 0] = -1
+                    ttsgn[i, xsi, 1] = -1
 
     else:
         tt[int(zsa), int(xsa)] = 0.0
 
+    # Start sweeping
     for _ in range(nsweep):
-        sweep2d(tt, ttgrad, slow, dz, dx, zsi, xsi, zsa, xsa, vzero, nz, nx, grad)
+        sweep2d(tt, ttsgn, slow, dz, dx, zsi, xsi, zsa, xsa, vzero, nz, nx, grad)
 
     if grad:
         for i in range(nz):
             for j in range(nx):
-                sgntz = int(ttgrad[i, j, 0])
+                sgntz = ttsgn[i, j, 0]
                 if sgntz != 0:
                     t1 = tt[i - sgntz, j]
                     ttgrad[i, j, 0] = sgntz * (tt[i, j] - t1) / dz
 
-                sgntx = int(ttgrad[i, j, 1])
+                sgntx = ttsgn[i, j, 1]
                 if sgntx != 0:
                     t1 = tt[i, j - sgntx]
                     ttgrad[i, j, 1] = sgntx * (tt[i, j] - t1) / dx
 
+                # Normalize gradients
+                gn = norm2d(ttgrad[i, j, 0], ttgrad[i, j, 1])
+                if gn > 0.0:
+                    ttgrad[i, j] /= gn
+
     return tt, ttgrad, vzero
 
 
