@@ -1,7 +1,7 @@
 import numpy
 from numba import prange
 
-from .._common import jitted
+from .._common import jitted, norm3d
 
 Big = 1.0e5
 eps = 1.0e-15
@@ -9,7 +9,7 @@ eps = 1.0e-15
 
 @jitted("f8(i4, i4, i4, f8, f8, f8, f8, f8, f8, f8)")
 def t_ana(i, j, k, dz, dx, dy, zsa, xsa, ysa, vzero):
-    """Calculate analytical times in homogenous model."""
+    """Calculate analytical times in homogeneous model."""
     return (
         vzero
         * ((dz * (i - zsa)) ** 2.0 + (dx * (j - xsa)) ** 2.0 + (dy * (k - ysa)) ** 2.0)
@@ -19,7 +19,7 @@ def t_ana(i, j, k, dz, dx, dy, zsa, xsa, ysa, vzero):
 
 @jitted("UniTuple(f8, 4)(i4, i4, i4, f8, f8, f8, f8, f8, f8, f8)")
 def t_anad(i, j, k, dz, dx, dy, zsa, xsa, ysa, vzero):
-    """Calculate analytical times in homogenous model and derivatives of times."""
+    """Calculate analytical times in homogeneous model and derivatives of times."""
     t = t_ana(i, j, k, dz, dx, dy, zsa, xsa, ysa, vzero)
 
     if t > 0.0:
@@ -36,20 +36,13 @@ def t_anad(i, j, k, dz, dx, dy, zsa, xsa, ysa, vzero):
 
 
 @jitted(
-    "void(f8[:, :, :], f8[:, :, :, :], f8[:, :, :], UniTuple(f8, 10), f8, f8, f8, f8, f8, f8, f8, i4, i4, i4, i4, i4, i4, i4, i4, i4, i4, i4, i4, b1)"
+    "void(f8[:, :, :], i4[:, :, :, :], f8[:, :, :], UniTuple(f8, 10), i4, i4, i4, i4, i4, i4, i4, i4, i4, i4, i4, i4, b1)"
 )
 def sweep(
     tt,
-    ttgrad,
+    ttsgn,
     slow,
     dargs,
-    zsi,
-    xsi,
-    ysi,
-    zsa,
-    xsa,
-    ysa,
-    vzero,
     i,
     j,
     k,
@@ -172,63 +165,46 @@ def sweep(
     # Compute gradient according to minimum time direction
     if grad and tt[i, j, k] != t0:
         if tt[i, j, k] == t1d1:
-            ttgrad[i, j, k, 0] = sgntz
-            ttgrad[i, j, k, 1] = 0.0
-            ttgrad[i, j, k, 2] = 0.0
+            ttsgn[i, j, k, 0] = sgntz
+            ttsgn[i, j, k, 1] = 0
+            ttsgn[i, j, k, 2] = 0
 
         elif tt[i, j, k] == t1d2:
-            ttgrad[i, j, k, 0] = 0.0
-            ttgrad[i, j, k, 1] = sgntx
-            ttgrad[i, j, k, 2] = 0.0
+            ttsgn[i, j, k, 0] = 0
+            ttsgn[i, j, k, 1] = sgntx
+            ttsgn[i, j, k, 2] = 0
 
         elif tt[i, j, k] == t1d3:
-            ttgrad[i, j, k, 0] = 0.0
-            ttgrad[i, j, k, 1] = 0.0
-            ttgrad[i, j, k, 2] = sgnty
+            ttsgn[i, j, k, 0] = 0
+            ttsgn[i, j, k, 1] = 0
+            ttsgn[i, j, k, 2] = sgnty
 
         elif tt[i, j, k] == t2d1:
-            ttgrad[i, j, k, 0] = sgntz
-            ttgrad[i, j, k, 1] = sgntx
-            ttgrad[i, j, k, 2] = 0.0
+            ttsgn[i, j, k, 0] = sgntz
+            ttsgn[i, j, k, 1] = sgntx
+            ttsgn[i, j, k, 2] = 0
 
         elif tt[i, j, k] == t2d2:
-            ttgrad[i, j, k, 0] = sgntz
-            ttgrad[i, j, k, 1] = 0.0
-            ttgrad[i, j, k, 2] = sgnty
+            ttsgn[i, j, k, 0] = sgntz
+            ttsgn[i, j, k, 1] = 0
+            ttsgn[i, j, k, 2] = sgnty
 
         elif tt[i, j, k] == t2d3:
-            ttgrad[i, j, k, 0] = 0.0
-            ttgrad[i, j, k, 1] = sgntx
-            ttgrad[i, j, k, 2] = sgnty
+            ttsgn[i, j, k, 0] = 0
+            ttsgn[i, j, k, 1] = sgntx
+            ttsgn[i, j, k, 2] = sgnty
 
         else:
-            ttgrad[i, j, k, 0] = sgntz
-            ttgrad[i, j, k, 1] = sgntx
-            ttgrad[i, j, k, 2] = sgnty
+            ttsgn[i, j, k, 0] = sgntz
+            ttsgn[i, j, k, 1] = sgntx
+            ttsgn[i, j, k, 2] = sgnty
 
 
 @jitted(
-    "void(f8[:, :, :], f8[:, :, :, :], f8[:, :, :], f8, f8, f8, f8, f8, f8, f8, f8, f8, f8, i4, i4, i4, b1, b1)"
+    "void(f8[:, :, :], i4[:, :, :, :], f8[:, :, :], f8, f8, f8, f8, f8, f8, i4, i4, i4, b1, b1)"
 )
 def sweep3d(
-    tt,
-    ttgrad,
-    slow,
-    dz,
-    dx,
-    dy,
-    zsi,
-    xsi,
-    ysi,
-    zsa,
-    xsa,
-    ysa,
-    vzero,
-    nz,
-    nx,
-    ny,
-    grad,
-    init=False,
+    tt, ttsgn, slow, dz, dx, dy, zsi, xsi, ysi, nz, nx, ny, grad, init=False,
 ):
     """Perform one sweeping."""
     dz2i = 1.0 / dz / dz
@@ -250,30 +226,7 @@ def sweep3d(
         for j in range(j0, nx):
             for i in range(i0, nz):
                 sweep(
-                    tt,
-                    ttgrad,
-                    slow,
-                    dargs,
-                    zsi,
-                    xsi,
-                    ysi,
-                    zsa,
-                    xsa,
-                    ysa,
-                    vzero,
-                    i,
-                    j,
-                    k,
-                    1,
-                    1,
-                    1,
-                    1,
-                    1,
-                    1,
-                    nz,
-                    nx,
-                    ny,
-                    grad,
+                    tt, ttsgn, slow, dargs, i, j, k, 1, 1, 1, 1, 1, 1, nz, nx, ny, grad,
                 )
 
     # Second sweeping: Top -> Bottom; East -> West; South -> North
@@ -287,16 +240,9 @@ def sweep3d(
             for i in range(i0, nz):
                 sweep(
                     tt,
-                    ttgrad,
+                    ttsgn,
                     slow,
                     dargs,
-                    zsi,
-                    xsi,
-                    ysi,
-                    zsa,
-                    xsa,
-                    ysa,
-                    vzero,
                     i,
                     j,
                     k,
@@ -323,16 +269,9 @@ def sweep3d(
             for i in range(i0, nz):
                 sweep(
                     tt,
-                    ttgrad,
+                    ttsgn,
                     slow,
                     dargs,
-                    zsi,
-                    xsi,
-                    ysi,
-                    zsa,
-                    xsa,
-                    ysa,
-                    vzero,
                     i,
                     j,
                     k,
@@ -359,16 +298,9 @@ def sweep3d(
             for i in range(i0, nz):
                 sweep(
                     tt,
-                    ttgrad,
+                    ttsgn,
                     slow,
                     dargs,
-                    zsi,
-                    xsi,
-                    ysi,
-                    zsa,
-                    xsa,
-                    ysa,
-                    vzero,
                     i,
                     j,
                     k,
@@ -395,16 +327,9 @@ def sweep3d(
             for i in range(i0, -1, -1):
                 sweep(
                     tt,
-                    ttgrad,
+                    ttsgn,
                     slow,
                     dargs,
-                    zsi,
-                    xsi,
-                    ysi,
-                    zsa,
-                    xsa,
-                    ysa,
-                    vzero,
                     i,
                     j,
                     k,
@@ -430,16 +355,9 @@ def sweep3d(
             for i in range(i0, -1, -1):
                 sweep(
                     tt,
-                    ttgrad,
+                    ttsgn,
                     slow,
                     dargs,
-                    zsi,
-                    xsi,
-                    ysi,
-                    zsa,
-                    xsa,
-                    ysa,
-                    vzero,
                     i,
                     j,
                     k,
@@ -466,16 +384,9 @@ def sweep3d(
             for i in range(i0, -1, -1):
                 sweep(
                     tt,
-                    ttgrad,
+                    ttsgn,
                     slow,
                     dargs,
-                    zsi,
-                    xsi,
-                    ysi,
-                    zsa,
-                    xsa,
-                    ysa,
-                    vzero,
                     i,
                     j,
                     k,
@@ -502,16 +413,9 @@ def sweep3d(
             for i in range(i0, -1, -1):
                 sweep(
                     tt,
-                    ttgrad,
+                    ttsgn,
                     slow,
                     dargs,
-                    zsi,
-                    xsi,
-                    ysi,
-                    zsa,
-                    xsa,
-                    ysa,
-                    vzero,
                     i,
                     j,
                     k,
@@ -537,9 +441,9 @@ def fteik3d(slow, dz, dx, dy, zsrc, xsrc, ysrc, nsweep=2, grad=False):
     nz, nx, ny = numpy.shape(slow)
 
     # Check inputs
-    condz = 0.0 <= zsrc <= dz * (nz - 1)
-    condx = 0.0 <= xsrc <= dx * (nx - 1)
-    condy = 0.0 <= ysrc <= dy * (ny - 1)
+    condz = 0.0 <= zsrc <= dz * nz
+    condx = 0.0 <= xsrc <= dx * nx
+    condy = 0.0 <= ysrc <= dy * ny
     if not (condz and condx and condy):
         raise ValueError("source out of bound")
 
@@ -564,11 +468,14 @@ def fteik3d(slow, dz, dx, dy, zsrc, xsrc, ysrc, nsweep=2, grad=False):
     nx += 1
     ny += 1
     tt = numpy.full((nz, nx, ny), Big, dtype=numpy.float64)
-    ttgrad = (
-        numpy.zeros((nz, nx, ny, 3), dtype=numpy.float64)
-        if grad
-        else numpy.empty((0, 0, 0, 0), dtype=numpy.float64)
-    )
+
+    if grad:
+        ttgrad = numpy.zeros((nz, nx, ny, 3), dtype=numpy.float64)
+        ttsgn = numpy.zeros((nz, nx, ny, 3), dtype=numpy.int32)
+
+    else:
+        ttgrad = numpy.empty((0, 0, 0, 0), dtype=numpy.float64)
+        ttsgn = numpy.empty((0, 0, 0, 0), dtype=numpy.int32)
 
     # Initialize points around source
     iterables = (
@@ -589,47 +496,38 @@ def fteik3d(slow, dz, dx, dy, zsrc, xsrc, ysrc, nsweep=2, grad=False):
             ttgrad[i, j, k, 1] = txc
             ttgrad[i, j, k, 2] = tyc
 
+    # Start sweeping
     for i in range(nsweep):
         sweep3d(
-            tt,
-            ttgrad,
-            slow,
-            dz,
-            dx,
-            dy,
-            zsi,
-            xsi,
-            ysi,
-            zsa,
-            xsa,
-            ysa,
-            vzero,
-            nz,
-            nx,
-            ny,
-            grad,
-            i == 0,
+            tt, ttsgn, slow, dz, dx, dy, zsi, xsi, ysi, nz, nx, ny, grad, i == 0,
         )
 
     if grad:
         for i in range(nz):
             for j in range(nx):
                 for k in range(ny):
-                    sgntz = int(ttgrad[i, j, k, 0])
+                    sgntz = ttsgn[i, j, k, 0]
                     if sgntz != 0:
                         t1 = tt[i - sgntz, j, k]
                         ttgrad[i, j, k, 0] = sgntz * (tt[i, j, k] - t1) / dz
 
-                    sgntx = int(ttgrad[i, j, k, 1])
+                    sgntx = ttsgn[i, j, k, 1]
                     if sgntx != 0:
                         t1 = tt[i, j - sgntx, k]
                         ttgrad[i, j, k, 1] = sgntx * (tt[i, j, k] - t1) / dx
 
-                    sgnty = int(ttgrad[i, j, k, 2])
+                    sgnty = ttsgn[i, j, k, 2]
                     if sgnty != 0:
                         t1 = tt[i, j, k - sgnty]
                         ttgrad[i, j, k, 2] = sgnty * (tt[i, j, k] - t1) / dy
 
+                    # Normalize gradients
+                    gn = norm3d(
+                        ttgrad[i, j, k, 0], ttgrad[i, j, k, 1], ttgrad[i, j, k, 2]
+                    )
+                    if gn > 0.0:
+                        ttgrad[i, j, k] /= gn
+
     return tt, ttgrad, vzero
 
 
