@@ -1,9 +1,11 @@
-#%%
+from copy import deepcopy
 
 import numpy
 import pytest
 from helpers import allclose, eik2d
 
+eik2d_copy = deepcopy(eik2d)
+
 
 @pytest.mark.parametrize(
     "sources, tref",
@@ -29,3 +31,18 @@ def test_solve(sources, tref):
 )
 def test_call(points, vref):
     allclose(vref, points, lambda x: eik2d(x))
+
+
+def test_resample():
+    eik2d = deepcopy(eik2d_copy)
+    nz, nx = eik2d.shape
+    eik2d.resample((nz * 2, nx * 2))
+
+    assert eik2d.grid.sum() == nz * nx * 4.0
+
+
+def test_smooth():
+    eik2d = deepcopy(eik2d_copy)
+    eik2d.smooth(1.0)
+
+    assert eik2d.grid.sum() == eik2d.grid.size
